@@ -5,6 +5,7 @@
 #include <ctype.h>
 #include <getopt.h>
 #include <errno.h>
+#include <fcntl.h>
 #include <sched.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -12,6 +13,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <sys/stat.h>
 #include <sys/types.h>
 
 #include "options.h"
@@ -38,17 +40,22 @@ const struct option_info options_info[] = {
   { C_L, OPT_LOCKOPT_QUIET,'x', NULL,       no_argument,       "fail silently" },
   { C_RS,OPT_LIMIT_MEM,   'm',  NULL,       required_argument,
     "set soft DATA, STACK, MEMLOCK and AS limits", "BYTES" },
-  { C_RS,OPT_RLIMIT_DATA, 'd',  NULL,       required_argument, "set soft RLIMIT_DATA", "BYTES" },
-  { C_XS,OPT_RLIMIT_STACK,'s',  NULL,       required_argument, "set soft RLIMIT_STACK", "BYTES" },
-  { C_S, OPT_RLIMIT_MEMLOCK,'l', NULL,      required_argument, "set soft RLIMIT_MEMLOCK", "BYTES" },
-  { C_X, OPT_RLIMIT_MEMLOCK,'\0',"memlock", required_argument, "set soft RLIMIT_MEMLOCK", "BYTES" },
-  { C_XS,OPT_RLIMIT_AS,   'a',  NULL,       required_argument, "set soft RLIMIT_AS", "BYTES" },
-  { C_RS,OPT_RLIMIT_NOFILE,'o', NULL,       required_argument, "set soft RLIMIT_NOFILE", "FILES" },
-  { C_RS,OPT_RLIMIT_NPROC,'p',  NULL,       required_argument, "set soft RLIMIT_NPROC", "PROCS" },
-  { C_RS,OPT_RLIMIT_FSIZE,'f',  NULL,       required_argument, "set soft RLIMIT_FSIZE", "BYTES" },
-  { C_RS,OPT_RLIMIT_CORE, 'c',  NULL,       required_argument, "set soft RLIMIT_CORE", "BYTES" },
-  { C_XS,OPT_RLIMIT_RSS,  'r',  NULL,       required_argument, "set soft RLIMIT_RSS", "BYTES" },
-  { C_RS,OPT_RLIMIT_CPU,  't',  NULL,       required_argument, "set soft RLIMIT_CPU", "SECONDS" },
+  { C_RS,OPT_RLIMIT_DATA, 'd',  NULL,       required_argument, "set RLIMIT_DATA", "BYTES" },
+  { C_XS,OPT_RLIMIT_STACK,'s',  NULL,       required_argument, "set RLIMIT_STACK", "BYTES" },
+  { C_S, OPT_RLIMIT_MEMLOCK,'l', NULL,      required_argument, "set RLIMIT_MEMLOCK", "BYTES" },
+  { C_XS,OPT_RLIMIT_AS,   'a',  NULL,       required_argument, "set RLIMIT_AS", "BYTES" },
+  { C_RS,OPT_RLIMIT_NOFILE,'o', NULL,       required_argument, "set RLIMIT_NOFILE", "FILES" },
+  { C_RS,OPT_RLIMIT_NPROC,'p',  NULL,       required_argument, "set RLIMIT_NPROC", "PROCS" },
+  { C_RS,OPT_RLIMIT_FSIZE,'f',  NULL,       required_argument, "set RLIMIT_FSIZE", "BYTES" },
+  { C_RS,OPT_RLIMIT_CORE, 'c',  NULL,       required_argument, "set RLIMIT_CORE", "BYTES" },
+  { C_XS,OPT_RLIMIT_RSS,  'r',  NULL,       required_argument, "set RLIMIT_RSS", "BYTES" },
+  { C_RS,OPT_RLIMIT_CPU,  't',  NULL,       required_argument, "set RLIMIT_CPU", "SECONDS" },
+  { C_X, OPT_RLIMIT_MEMLOCK,'\0',"limit-memlock", required_argument, "set RLIMIT_MEMLOCK", "BYTES" },
+  { C_X, OPT_RLIMIT_MSGQUEUE, '\0', "limit-msgqueue", required_argument, "set RLIMIT_MSGQUEUE", "BYTES" },
+  { C_X, OPT_RLIMIT_NICE, '\0', "limit-nice", required_argument, "set RLIMIT_NICE", "NICENESS" },
+  { C_X, OPT_RLIMIT_RTPRIO, '\0', "limit-rtprio", required_argument, "set RLIMIT_RTPRIO", "PRIO" },
+  { C_X, OPT_RLIMIT_RTTIME, '\0', "limit-rttime", required_argument, "set RLIMIT_RTTIME", "MS" },
+  { C_X, OPT_RLIMIT_SIGPENDING, '\0', "limit-sigpending", required_argument, "set RLIMIT_SIGPENDING", "NUM" },
   { C_ALL,OPT_VERBOSE,    'v',  "verbose",  no_argument,       "be verbose" },
   { C_R, OPT_VERSION,     'V',  "version",  no_argument,       "show " NAME_STR " version" },
   { C_R, OPT_PGRPHACK,    'P',  NULL,       no_argument,       "run in new process group" },
@@ -58,6 +65,7 @@ const struct option_info options_info[] = {
   { C_X, OPT_LEGACY,      '@',  NULL,       no_argument,
     "restricts following options to chpst(8) ones" },
   { C_X, OPT_HELP,        'h',  "help",     no_argument,       "show help" },
+  { C_X, OPT_FILE,        '\0', "file",     required_argument, "read options from file", "FILE" },
   { C_X, OPT_EXIT,        '\0', "exit",     optional_argument,
     "exit (with optional RETCODE)", "RETCODE" },
   { C_X, OPT_MOUNT_NS,    '\0', "mount-ns", no_argument,       "create mount namespace" },
@@ -82,9 +90,10 @@ const struct option_info options_info[] = {
   { C_X, OPT_FORK_JOIN,   '\0', "fork-join", no_argument,      "fork and wait for process" },
   { C_X, OPT_NEW_ROOT,    '\0', "new-root",  no_argument,      "create a new root fs" },
   { C_X, OPT_NO_NEW_PRIVS,'\0', "no-new-privs", no_argument,   "no new privileges" },
-  { C_X, OPT_SCHEDULER,   '\0', "scheduler", required_argument,"set scheduler policy", "POLICY" },
   { C_X, OPT_CPUS,        '\0', "cpus",      required_argument,"set CPU affinity", "AFFINITY" },
-  { C_X, OPT_IO_NICE,     '\0', "io-nice",   required_argument,
+  { C_X, OPT_CPU_SCHED,   '\0', "cpu-scheduler", required_argument,
+    "set CPU scheduler policy", "POLICY" },
+  { C_X, OPT_IO_SCHED,    '\0', "io-scheduler",   required_argument,
     "set I/O scheduling class", "rt|best-effort|idle[:PRIORITY]"},
   { C_X, OPT_UMASK,       '\0', "umask",     required_argument,"set umask", "MODE" },
   { C_X, OPT_APP,         '\0', "app",       required_argument,"define application name", "NAME"},
@@ -93,12 +102,17 @@ const struct option_info options_info[] = {
   { C_X, OPT_CACHE_DIR,   '\0', "cache-dir", no_argument,      "create cache dir" },
   { C_X, OPT_LOG_DIR,     '\0', "log-dir",   no_argument,      "create log dir" },
   { C_X, OPT_LOGIN,       '\0', "login",     no_argument,      "simulate login environment" },
+  { C_X, OPT_OOM,         '\0', "oom",       required_argument,"set oom adjust value", "ADJ" },
 };
 #define max_options ((ssize_t) ((sizeof options_info / sizeof *options_info)))
 
 static struct option options[max_options + 1];
 static char optstr[max_options * 2];
 
+static void handle_option(enum compat_level *compat,
+                          const struct option_info *optdef,
+                          char *optarg);
+
 void options_print(FILE *out) {
   const struct option_info *optdef;
   int i = 0;
@@ -222,7 +236,9 @@ bool parse_limit(rlim_t *lim, const char *arg) {
   int toks = 0;
   long long val;
 
-  if (!strcmp(arg, "unlimited")) {
+  if (!strcmp(arg, "unlimited") ||
+      !strcmp(arg, "infinity") ||
+      !strcmp(arg, "=")) {
     *lim = RLIM_INFINITY;
     toks = 1;
   } else {
@@ -241,6 +257,35 @@ bool parse_limit(rlim_t *lim, const char *arg) {
   return toks == 1 ? true : false;
 }
 
+/* SUPPORTED FORMATS
+ *   soft    (this differs from prlimit to match chpst)
+ *   soft:
+ *   soft:hard
+ *   :hard
+ *   +both   (this is unique to xchpst)
+ */
+bool parse_limits(struct limit *limit, const char *arg) {
+  char *sep;
+  bool both = *arg == '+';
+
+  if (both)
+    arg++;
+
+  sep = strchr(arg, ':');
+  if (sep != NULL) {
+    *sep++ = '\0';
+    limit->hard_specified = parse_limit(&limit->limits.rlim_max, sep);
+  }
+  if (arg != sep) {
+    limit->soft_specified = parse_limit(&limit->limits.rlim_cur, arg);
+  }
+  if (both) {
+    limit->limits.rlim_max = limit->limits.rlim_cur;
+    limit->hard_specified = limit->soft_specified;
+  }
+  return limit->soft_specified || limit->hard_specified;
+}
+
 bool parse_caps(cap_bits_t *caps, char *names) {
   cap_bits_t set = 0;
   cap_value_t val;
@@ -373,8 +418,215 @@ int sched_policy_from_name(const char *name) {
   return SCHED_OTHER;
 }
 
+static const struct option_info *find_option(int by_code,
+                                             const char *by_name) {
+  const struct option_info *optdef;
+  const struct option_info *incompatible_option = NULL;
+  enum compat_level compat = opt.app->compat_level;
+  bool is_short = (by_code && isascii(by_code)) ||
+                  (by_name && *by_name && by_name[1] == '\0');
+
+  /* Look up option definition */
+  for (optdef = options_info;
+       optdef - options_info < max_options;
+       optdef++) {
+
+    if ((by_code && by_code == (is_short ? optdef->short_name : (int) optdef->option)) ||
+        (by_name && (is_short ? *by_name == optdef->short_name : optdef->long_name && !strcmp(by_name, optdef->long_name)))) {
+      if ((optdef->compat_level & compat) == 0) {
+        /* Tentatively found an incompatible option but there may be
+         * another one that is compatible later in the list. */
+        incompatible_option = optdef;
+      } else {
+        incompatible_option = NULL;
+        break;
+      }
+    }
+  }
+
+  if (optdef - options_info == max_options)
+    return incompatible_option;
+  else
+    return optdef;
+}
+
+static void read_options_file(const char *path) {
+  struct options_file *opt_file;
+  struct stat statbuf;
+  int fd;
+  int rc;
+  off_t offset;
+  char *ptr;
+  char *option_name = NULL;
+  char *option_name_end = NULL;
+  char *option_value = NULL;
+  char *option_value_end = NULL;
+  ssize_t len;
+  const struct option_info *optdef;
+  enum compat_level compat = opt.app->compat_level;
+  enum {
+    S_LEADING_WSP,
+    S_COMMENT,
+    /* No states above are actionable on EOF */
+    /* All states below are actionable on EOF */
+    S_KEY,
+    S_DELIM_WSP,
+    S_VALUE,
+    S_POSSIBLY_TRAILING_WSP,
+  } state = S_LEADING_WSP;
+
+  fd = open(path, O_RDONLY);
+  if (fd == -1) {
+    fprintf(stderr, "error opening options file %s: %s\n", path, strerror(errno));
+    goto fail;
+  }
+
+  rc = fstat(fd, &statbuf);
+  if (rc == -1) {
+    close(fd);
+    goto fail;
+  }
+
+  /* We allocate the space for the file permanently, just as the argv array
+     remains allocated. Option processing modifies argument values.
+     Add one for NUL-terminating last element if not a newline. */
+  opt_file = malloc(sizeof *opt_file + statbuf.st_size + 1);
+  if (opt_file == NULL) {
+    close(fd);
+    goto fail;
+  }
+
+  opt_file->next = opt.opt_files;
+  opt.opt_files = opt_file;
+
+  offset = 0;
+  do {
+    len = read(fd, opt_file->content + offset, statbuf.st_size - offset);
+    if (len == -1) {
+      close(fd);
+      goto fail;
+    }
+    offset += len;
+  } while (offset < statbuf.st_size);
+  close(fd);
+
+  ptr = opt_file->content;
+  for (ptr = opt_file->content; ptr - opt_file->content < statbuf.st_size; ptr++) {
+    char c = *ptr;
+
+    /* Whether a key and value are now completely specified */
+    bool action = false;
+
+    switch (state) {
+    case S_LEADING_WSP:
+      if (c == '#') {
+        state = S_COMMENT;
+      } else if (!isspace(c)) {
+        option_name = ptr;
+        state = S_KEY;
+      }
+      break;
+    case S_KEY:
+      if (c == '\n') {
+        action = true;
+      } else if (isspace(c)) {
+        state = S_DELIM_WSP;
+      }
+      option_name_end = ptr;
+      break;
+    case S_DELIM_WSP:
+      if (c == '\n') {
+        action = true;
+      } else if (!isspace(c)) {
+        option_value = ptr;
+        state = S_VALUE;
+      }
+      break;
+    case S_VALUE:
+      if (c == '\n') {
+        action = true;
+      } else if (isspace(c)) {
+        state = S_POSSIBLY_TRAILING_WSP;
+      }
+      option_value_end = ptr;
+      break;
+    case S_POSSIBLY_TRAILING_WSP:
+      if (c == '\n') {
+        action = true;
+      } else if (!isspace(c)) {
+        state = S_VALUE;
+        option_value_end = ptr;
+      }
+      break;
+    case S_COMMENT:
+      if (c == '\n') {
+        state = S_LEADING_WSP;
+      }
+      break;
+    }
+
+    if (action /* EOL */ ||
+        (ptr + 1 == opt_file->content + statbuf.st_size /* EOF */ &&
+         state >= S_KEY /* An actionable state */)) {
+      if (!action) {
+        /* If EOF then there is no EOL character to overwrite */
+        if (option_value)
+          option_value_end = ptr + 1;
+        else
+          option_name_end = ptr + 1;
+      }
+      *option_name_end = '\0';
+      if (option_value)
+        *option_value_end = '\0';
+
+      state = S_LEADING_WSP;
+    } else {
+      continue;
+    }
+
+    optdef = find_option(0, option_name);
+
+    if (optdef && (optdef->compat_level & compat) == 0) {
+      char short_name[2] = { optdef->short_name, 0};
+      fprintf(stderr, "illegal option (%s) at this compat level\n",
+              optdef->long_name ? optdef->long_name : short_name);
+      opt.error = true;
+      continue;
+    } else if (!optdef) {
+      fprintf(stderr, "unknown option in config file: %s\n", option_name);
+      opt.error = true;
+      continue;
+    }
+
+    if (is_verbose())
+      fprintf(stderr, "handling file option '%s' with value '%s'\n", option_name, option_value);
+
+    enable(optdef->option);
+
+    /* For now, throw away the value end pointer. In future we could use
+       this to enable arguments containing NUL bytes to be handled.
+       (Of course, this format cannot handle newline characters, which would
+       probably be more useful to embed!) */
+    handle_option(&compat, optdef,
+                  optdef->has_arg == no_argument ? NULL : option_value);
+    option_value = NULL;
+  }
+
+  if (errno != 0) {
+    fprintf(stderr, "error reading options file %s: %s\n", path, strerror(errno));
+    goto fail;
+  }
+
+  return;
+fail:
+   opt.error = true;
+}
+
 static void handle_option(enum compat_level *compat,
-                          const struct option_info *optdef) {
+                          const struct option_info *optdef,
+                          char *optarg) {
+  char *end;
+
   switch (optdef->option) {
   case OPT_LEGACY:
     *compat = COMPAT_CHPST;
@@ -385,6 +637,9 @@ static void handle_option(enum compat_level *compat,
   case OPT_HELP:
     opt.help = true;
     break;
+  case OPT_FILE:
+    read_options_file(optarg);
+    break;
   case OPT_EXIT:
     opt.exit = true;
     opt.retcode = optarg ? atoi(optarg) : CHPST_ERROR_EXIT;
@@ -479,19 +734,24 @@ static void handle_option(enum compat_level *compat,
       opt.error = true;
     opt.caps_op = CAP_OP_DROP;
     break;
-  case OPT_SCHEDULER:
+  case OPT_CPU_SCHED:
     opt.sched_policy = sched_policy_from_name(optarg);
     break;
   case OPT_CPUS:
     parse_cpus(optarg);
     break;
-  case OPT_IO_NICE:
+  case OPT_IO_SCHED:
     parse_ionice(optarg);
     break;
   case OPT_UMASK:
     if (sscanf(optarg, "%o", &opt.umask) != 1)
       opt.error = true;
     break;
+  case OPT_OOM:
+    opt.oom_adjust = strtol(optarg, &end, 10);
+    if (*optarg == '\0' || *end != '\0')
+      opt.error = true;
+    break;
   case OPT_APP:
     opt.app_name = optarg;
     break;
@@ -512,7 +772,7 @@ static void handle_option(enum compat_level *compat,
       usrgrp_print(stderr, "envuidgid", &opt.env_users_groups);
     break;
   case OPT_LIMIT_MEM:
-    if (!(opt.rlimit_memlock.soft_specified = parse_limit(&opt.rlimit_memlock.limits.rlim_cur, optarg))) {
+    if (!parse_limits(&opt.rlimit_memlock, optarg)) {
       opt.error = true;
     } else {
       opt.rlimit_data = opt.rlimit_memlock;
@@ -521,43 +781,63 @@ static void handle_option(enum compat_level *compat,
     }
     break;
    case OPT_RLIMIT_DATA:
-    if (!(opt.rlimit_data.soft_specified = parse_limit(&opt.rlimit_data.limits.rlim_cur, optarg)))
+    if (!parse_limits(&opt.rlimit_data, optarg))
       opt.error = true;
     break;
   case OPT_RLIMIT_MEMLOCK:
-    if (!(opt.rlimit_memlock.soft_specified = parse_limit(&opt.rlimit_memlock.limits.rlim_cur, optarg)))
+    if (!parse_limits(&opt.rlimit_memlock, optarg))
       opt.error = true;
     break;
   case OPT_RLIMIT_AS:
-    if (!(opt.rlimit_as.soft_specified = parse_limit(&opt.rlimit_as.limits.rlim_cur, optarg)))
+    if (!parse_limits(&opt.rlimit_as, optarg))
       opt.error = true;
     break;
   case OPT_RLIMIT_STACK:
-    if (!(opt.rlimit_stack.soft_specified = parse_limit(&opt.rlimit_stack.limits.rlim_cur, optarg)))
+    if (!parse_limits(&opt.rlimit_stack, optarg))
       opt.error = true;
     break;
   case OPT_RLIMIT_NOFILE:
-    if (!(opt.rlimit_nofile.soft_specified = parse_limit(&opt.rlimit_nofile.limits.rlim_cur, optarg)))
+    if (!parse_limits(&opt.rlimit_nofile, optarg))
       opt.error = true;
     break;
   case OPT_RLIMIT_RSS:
-    if (!(opt.rlimit_rss.soft_specified = parse_limit(&opt.rlimit_rss.limits.rlim_cur, optarg)))
+    if (!parse_limits(&opt.rlimit_rss, optarg))
       opt.error = true;
     break;
   case OPT_RLIMIT_NPROC:
-    if (!(opt.rlimit_nproc.soft_specified = parse_limit(&opt.rlimit_nproc.limits.rlim_cur, optarg)))
+    if (!parse_limits(&opt.rlimit_nproc, optarg))
       opt.error = true;
     break;
   case OPT_RLIMIT_FSIZE:
-   if (!(opt.rlimit_fsize.soft_specified = parse_limit(&opt.rlimit_fsize.limits.rlim_cur, optarg)))
+    if (!parse_limits(&opt.rlimit_fsize, optarg))
       opt.error = true;
     break;
   case OPT_RLIMIT_CPU:
-   if (!(opt.rlimit_cpu.soft_specified = parse_limit(&opt.rlimit_cpu.limits.rlim_cur, optarg)))
+    if (!parse_limits(&opt.rlimit_cpu, optarg))
       opt.error = true;
     break;
   case OPT_RLIMIT_CORE:
-   if (!(opt.rlimit_core.soft_specified = parse_limit(&opt.rlimit_core.limits.rlim_cur, optarg)))
+    if (!parse_limits(&opt.rlimit_core, optarg))
+      opt.error = true;
+    break;
+  case OPT_RLIMIT_MSGQUEUE:
+    if (!parse_limits(&opt.rlimit_msgqueue, optarg))
+      opt.error = true;
+    break;
+  case OPT_RLIMIT_NICE:
+    if (!parse_limits(&opt.rlimit_nice, optarg))
+      opt.error = true;
+    break;
+  case OPT_RLIMIT_RTPRIO:
+    if (!parse_limits(&opt.rlimit_rtprio, optarg))
+      opt.error = true;
+    break;
+  case OPT_RLIMIT_RTTIME:
+    if (!parse_limits(&opt.rlimit_rttime, optarg))
+      opt.error = true;
+    break;
+  case OPT_RLIMIT_SIGPENDING:
+    if (!parse_limits(&opt.rlimit_sigpending, optarg))
       opt.error = true;
     break;
   case OPT_CLOSE_STDIN:
@@ -580,45 +860,29 @@ int options_parse(int argc, char *argv[]) {
   enum compat_level compat = opt.app->compat_level;
   int c;
 
+  /* Process options */
   while ((c = opt.app->long_opts ?
                 getopt_long(argc, argv, optstr, options, NULL) :
                 getopt(argc, argv, optstr)) != -1) {
-    bool is_short = isascii(c);
-    const struct option_info *incompatible_option = NULL;
 
-    /* Look up option definition */
-    for (optdef = options_info;
-         optdef - options_info < max_options;
-         optdef++) {
+    optdef = find_option(c, NULL);
 
-      if (c == (is_short ? optdef->short_name : (int) optdef->option)) {
-        if ((optdef->compat_level & compat) == 0) {
-          /* Tentatively found an incompatible option but there may be
-           * another one that is compatible later in the list. */
-          incompatible_option = optdef;
-        } else {
-          incompatible_option = NULL;
-          break;
-        }
-      }
-    }
-
-    /* Handle option */
-    if (optdef - options_info == max_options) {
-      if ((optdef = incompatible_option)) {
-        char short_name[2] = { optdef->short_name, 0};
-        fprintf(stderr, "illegal option (%s) at this compat level\n",
-                optdef->long_name ? optdef->long_name : short_name);
-      }
+    if (optdef && (optdef->compat_level & compat) == 0) {
+      char short_name[2] = { optdef->short_name, 0};
+      fprintf(stderr, "illegal option (%s) at this compat level\n",
+              optdef->long_name ? optdef->long_name : short_name);
       opt.error = true;
-    } else {
+    } else if (optdef) {
       /* Set bitfield before calling handler in case the handler wishes
        * to reset the option based on argument value. */
       enable(optdef->option);
-      handle_option(&compat, optdef);
+      handle_option(&compat, optdef, optarg);
+    } else {
+      opt.error = true;
     }
   }
 
+  /* Process positional arguments */
   assert(opt.app->takes_positional_opts <= MAX_POSITIONAL_OPTS);
   for (int i = 0; i < opt.app->takes_positional_opts; i++) {
     enum opt option = opt.app->positional_opts[i];
@@ -633,16 +897,22 @@ int options_parse(int argc, char *argv[]) {
          optdef++);
     assert(optdef);
 
-    if (optdef->has_arg != no_argument)
-      optarg = argv[optind++];
-    handle_option(&compat, optdef);
+    handle_option(&compat, optdef,
+                  optdef->has_arg == no_argument ? NULL : argv[optind++]);
   }
   return optind;
 }
 
 void options_free(void) {
+  struct options_file *file, *next;
+
   if (opt.cpu_affinity.size)
     CPU_FREE(opt.cpu_affinity.mask);
   usrgrp_free(&opt.users_groups);
   usrgrp_free(&opt.env_users_groups);
+
+  for (file = opt.opt_files; file; file = next) {
+    next = file->next;
+    free(file);
+  }
 }
