@@ -162,22 +162,32 @@ fail0:
 void set_rlimit(int resource, struct limit *option) {
   struct rlimit prev;
 
-  if (option->soft_specified) {
-    if (getrlimit(resource, &prev) == 0) {
-      if (option->limits.rlim_cur != RLIM_INFINITY &&
-          option->limits.rlim_cur > prev.rlim_max) {
-        fprintf(stderr, "warning capping requested %d soft limit from %lld to maximum %lld\n",
-                resource, (long long) option->limits.rlim_cur, (long long) prev.rlim_max);
-        prev.rlim_cur = prev.rlim_max;
-      } else {
-        prev.rlim_cur = option->limits.rlim_cur;
-      }
-      if (setrlimit(resource, &prev) != 0) {
-        fprintf(stderr, "warning: failed to set %d soft limit\n", resource);
-      }
-    } else {
+  if (option->soft_specified || option->hard_specified) {
+    if (getrlimit(resource, &prev) != 0) {
       fprintf(stderr, "warning: resource type %d cannot be controlled on this kernel\n", resource);
+      return;
     }
+  } else {
+    return;
+  }
+
+  if (option->hard_specified)
+    prev.rlim_max = option->limits.rlim_max;
+
+  if (option->soft_specified) {
+    if (option->limits.rlim_cur != RLIM_INFINITY &&
+        option->limits.rlim_max != RLIM_INFINITY &&
+        option->limits.rlim_cur > prev.rlim_max) {
+      fprintf(stderr, "warning capping requested %d soft limit from %lld to maximum %lld\n",
+              resource, (long long) option->limits.rlim_cur, (long long) prev.rlim_max);
+      prev.rlim_cur = prev.rlim_max;
+    } else {
+      prev.rlim_cur = option->limits.rlim_cur;
+    }
+  }
+
+  if (setrlimit(resource, &prev) != 0) {
+    fprintf(stderr, "warning: failed to set type %d soft limit\n", resource);
   }
 }
 
@@ -192,6 +202,11 @@ void set_resource_limits(void) {
   set_rlimit(RLIMIT_FSIZE, &opt.rlimit_fsize);
   set_rlimit(RLIMIT_CORE, &opt.rlimit_core);
   set_rlimit(RLIMIT_CPU, &opt.rlimit_cpu);
+  set_rlimit(RLIMIT_MSGQUEUE, &opt.rlimit_msgqueue);
+  set_rlimit(RLIMIT_NICE, &opt.rlimit_nice);
+  set_rlimit(RLIMIT_RTPRIO, &opt.rlimit_rtprio);
+  set_rlimit(RLIMIT_RTTIME, &opt.rlimit_rttime);
+  set_rlimit(RLIMIT_SIGPENDING, &opt.rlimit_sigpending);
 }
 
 static const struct app *find_app(const char *name) {
@@ -287,6 +302,9 @@ int main(int argc, char *argv[]) {
   if (set(OPT_UMASK))
     umask(opt.umask);
 
+  if (set(OPT_OOM))
+    write_once("/proc/self/oom_score_adj", "%ld", opt.oom_adjust);
+
   if (opt.lock_file) {
     if (opt.lock_nowait_override)
       opt.lock_wait = false;
@@ -351,7 +369,7 @@ int main(int argc, char *argv[]) {
     }
   }
 
-  if (set(OPT_IO_NICE)) {
+  if (set(OPT_IO_SCHED)) {
     if (syscall(SYS_ioprio_set,IOPRIO_WHO_PROCESS, 0, opt.ionice_prio) == -1) {
       fprintf(stderr, "warning: failed to set I/O scheduling class\n");
     } else if (is_verbose()) {
@@ -366,7 +384,7 @@ int main(int argc, char *argv[]) {
                         opt.cpu_affinity.mask) == -1)
     perror("could not set CPU affinity");
 
-  if (set(OPT_SCHEDULER) &&
+  if (set(OPT_CPU_SCHED) &&
       sched_setscheduler(0, opt.sched_policy, &((struct sched_param) {})) == -1)
     perror("could not change scheduler policy");
 
